﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.IO;

//All the application logic is here
namespace Halma
{
    /// <summary>
    /// Interaction logic for Window1.xaml
    /// It includes Board initialization
    /// </summary>
    public partial class Window1 : Window
    {
        public static Board boardClass;
        public static Canvas boardCanvasStatic;
        public static string game_id = String.Empty;

        public Window1()
        {
            //XAML
            InitializeComponent();
            //BOARD
            boardCanvasStatic = boardCanvas;
            boardClass = new Board();
            boardClass.Init();
            //First Render
            boardClass.Render(boardCanvasStatic);

            game_id = DateTime.Now.ToString().Replace(". ", "-").Replace(" ","_").Replace(":","-") + ".txt";
        }

        //!TODO! - Implement rotation for player
        public static int rotateCount = 0;
        void MoveClick(object sender, RoutedEventArgs e)
        {
            //boardClass.Move(moveFrom.Text, moveTo.Text);
            //rotateCount++;
            //boardCanvasStatic.RenderTransform = //new RotateTransform(60.0 * rotateCount, boardCanvasStatic.Width/2.0, boardCanvasStatic.Height/2.0);
            //new RotateTransform(60.0 * rotateCount, 0, 0);

            //boardClass.Render(boardCanvasStatic);
        }

        //Cancel the first click on a field
        void CancelClick(object sender, RoutedEventArgs e)
        {
            clickFirst = "#";
            boardClass.Render(boardCanvasStatic);
        }

        //if the possible moves higlight is turned on
        public static bool highlight = false;
        void HighlightClick(object sender, RoutedEventArgs e)
        {
            highlight = !highlight;
            //if(highlight) Highlight.Da
        }

        //if you click, then the whole game history will be saved in a file
        void SaveClick(object sender, RoutedEventArgs e)
        {
            boardClass.Save("save.txt");
        }
        //if you click, then the whole game history will be saved in a file
        void ReplayClick(object sender, RoutedEventArgs e)
        {
            boardClass.Replay("save.txt");
        }
        
        //Rendering function with makes a circle around a field (you specify where and the color)
        static public void addHighlightEllipse(Point position, Brush brush)
        {
            Ellipse myEllipse = new Ellipse();
            myEllipse.StrokeThickness = 4;
            myEllipse.Stroke = brush;
            // Set the width and height of the Ellipse.
            myEllipse.Width = Board.DIST / 1.2 + 8;
            myEllipse.Height = Board.DIST / 1.2 + 8;
            Point hex_position = Board.XYtoHex(position);

            // Add the Ellipse to the StackPanel.
            Canvas.SetLeft(myEllipse, hex_position.X - 4);
            Canvas.SetBottom(myEllipse, hex_position.Y - 4);

            Window1.boardCanvasStatic.Children.Add(myEllipse);

        }

        //Handling the two-click design for the move input 
        static string clickFirst = "#";
        static public void myEllipse_MouseDown(object sender, MouseButtonEventArgs e)
        {
            
            if (sender is Ellipse)
            {
                string name = (sender as Ellipse).Name;
                //MessageBox.Show(name);

                if (clickFirst == "#")
                {
                    clickFirst = name;
                    addHighlightEllipse(Board.STRtoXY(clickFirst), Brushes.GreenYellow);

                    if (highlight)
                    {
                        HashSet<Point> possible_moves = boardClass.GetPossibleMoves(Board.STRtoXY(clickFirst));
                        foreach (Point position in possible_moves)
                        {
                            addHighlightEllipse(position, Brushes.Red);
                        }
                    }
                }
                else if (clickFirst != "#")
                {
                    boardClass.Move(clickFirst, name);
                    boardClass.Save(game_id);
                    boardClass.Render(boardCanvasStatic);
                    clickFirst = "#";
                }
            }
            
        }
           
    }

    //Stores and modify the board position using a Dictionary (Board.board)
    //I use five representations of a field position:
    //-Point -> X coordinate id (-left, +right) Y coordinate (-left bottom, +right up)
    //-String - {'A'-'Z'}{'0'-'20'} for example D2 stands for Point(3,2)
    //-Hex    - the real position of the field on the drawing canvas
    //-Ellipse - the name of the Ellipse is the string identifier (I include an event -> click)
    //-TextBox - so the players could identify the fields visually by the String representation
    //I use a Dictionary for all fields for the board (key->string representation, value->FieldType)
    //So on a move I just swap the contents in the Dictionary identified by the string representation of a position
    public class Board
    {
        public enum FieldType
        {
            Empty, Player1, Player2, Player3, Player4, Player5, Player6
        }
        //Distance between two fields
        public const int DIST = 40;
        public Dictionary<string, FieldType> board;
        public List<KeyValuePair<Point, Point>> history;

        public Board()
        {
            this.board = new Dictionary<string, FieldType>();
            this.history = new List<KeyValuePair<Point, Point>>();
        }

        //Set up all fields. It is done manually foreach player and the empty fields are filled.
        public void Init()
        {
            this.board.Add("A12", FieldType.Player1);
            this.board.Add("B12", FieldType.Player1);
            this.board.Add("B11", FieldType.Player1);
            this.board.Add("C12", FieldType.Player1);
            this.board.Add("C11", FieldType.Player1);
            this.board.Add("C10", FieldType.Player1);
            this.board.Add("D12", FieldType.Player1);
            this.board.Add("D11", FieldType.Player1);
            this.board.Add("D10", FieldType.Player1);
            this.board.Add("D9", FieldType.Player1);

            this.board.Add("E13", FieldType.Player2);
            this.board.Add("E14", FieldType.Player2);
            this.board.Add("E15", FieldType.Player2);
            this.board.Add("E16", FieldType.Player2);
            this.board.Add("F13", FieldType.Player2);
            this.board.Add("F14", FieldType.Player2);
            this.board.Add("F15", FieldType.Player2);
            this.board.Add("G13", FieldType.Player2);
            this.board.Add("G14", FieldType.Player2);
            this.board.Add("H13", FieldType.Player2);

            this.board.Add("J12", FieldType.Player3);
            this.board.Add("K12", FieldType.Player3);
            this.board.Add("K11", FieldType.Player3);
            this.board.Add("L12", FieldType.Player3);
            this.board.Add("L11", FieldType.Player3);
            this.board.Add("L10", FieldType.Player3);
            this.board.Add("M12", FieldType.Player3);
            this.board.Add("M11", FieldType.Player3);
            this.board.Add("M10", FieldType.Player3);
            this.board.Add("M9", FieldType.Player3);

            this.board.Add("Q4", FieldType.Player4);
            this.board.Add("P4", FieldType.Player4);
            this.board.Add("P5", FieldType.Player4);
            this.board.Add("O4", FieldType.Player4);
            this.board.Add("O5", FieldType.Player4);
            this.board.Add("O6", FieldType.Player4);
            this.board.Add("N4", FieldType.Player4);
            this.board.Add("N5", FieldType.Player4);
            this.board.Add("N6", FieldType.Player4);
            this.board.Add("N7", FieldType.Player4);

            this.board.Add("M0", FieldType.Player5);
            this.board.Add("M1", FieldType.Player5);
            this.board.Add("L1", FieldType.Player5);
            this.board.Add("M2", FieldType.Player5);
            this.board.Add("L2", FieldType.Player5);
            this.board.Add("K2", FieldType.Player5);
            this.board.Add("M3", FieldType.Player5);
            this.board.Add("L3", FieldType.Player5);
            this.board.Add("K3", FieldType.Player5);
            this.board.Add("J3", FieldType.Player5);

            this.board.Add("H4", FieldType.Player6);
            this.board.Add("G4", FieldType.Player6);
            this.board.Add("G5", FieldType.Player6);
            this.board.Add("F4", FieldType.Player6);
            this.board.Add("F5", FieldType.Player6);
            this.board.Add("F6", FieldType.Player6);
            this.board.Add("E4", FieldType.Player6);
            this.board.Add("E5", FieldType.Player6);
            this.board.Add("E6", FieldType.Player6);
            this.board.Add("E7", FieldType.Player6);

            //Goes thru a parallelopiped and fills the fields empty (which are not player fields)
            for (int i = 0; i < 9; i++) {
                for (int j = 0; j < 9; j++) {
                    string str = Board.XYtoSTR(new Point(4 + i, 4 + j));
                    if (this.board.ContainsKey(str)) continue;

                    this.board.Add(str, FieldType.Empty);
                }
            }
        }

        public void Move(string from, string to)
        {
            this.Move(from, to, true);
        }

        //Checks if such positions exist and if yes the swaps the content of the map
        public void Move(string from, string to, bool add_history)
        {
            //If no such positions exists, then nothing to do.
            if (!this.board.ContainsKey(from) || !this.board.ContainsKey(to)) return;

            FieldType type = this.board[from];
            this.board[from] = this.board[to];
            this.board[to] = type;

            //I`m saving the history
            if (add_history)
            {
                this.history.Add(new KeyValuePair<Point, Point>(Board.STRtoXY(from), Board.STRtoXY(to)));
            }
        }

        public void Save()
        {
          Save("save.txt");
        } 
        public void Save(string filename)
        {
            string output = String.Empty;
            foreach (KeyValuePair<Point, Point> move in this.history)
            {
                output += move.Key.X + "," + move.Key.Y +","+ move.Value.X + "," + move.Value.Y + "\n";
            }
            File.WriteAllText(filename, output);
        }


        //replay uses the history variable
        public System.Windows.Threading.DispatcherTimer replay_timer;
        public void Replay(string filename)
        {
            //MessageBox.Show("Replay.Start");
            replay_timer = new System.Windows.Threading.DispatcherTimer();

            history.Clear();
            string[] input = File.ReadAllLines(filename);
            foreach (string line in input)
            {
                string[] numbers = line.Split(',');
                history.Add(new KeyValuePair<Point,Point>(new Point(
                    Int32.Parse(numbers[0]), Int32.Parse(numbers[1]) 
                    ),  new Point(
                    Int32.Parse(numbers[2]), Int32.Parse(numbers[3]) 
                    )
                ));
            }
            //MessageBox.Show("Replay.History.Count = " + history.Count.ToString());
            replay_tick = 0;
            replay_timer.Interval = new TimeSpan(0, 0, 2);
            replay_timer.Tick += new EventHandler(Replaytimer_Tick);
            replay_timer.Start();
        }
        public int replay_tick;
        //!TODO! - Something is not working
        void Replaytimer_Tick(object sender, EventArgs e)
        {
            //MessageBox.Show("Replay.Tick");
            if (replay_tick >= history.Count)
            {
                replay_timer.Stop();
                return;
            }
            this.Move(Board.XYtoSTR(history[replay_tick].Key),Board.XYtoSTR(history[replay_tick].Value),
                false);
            this.Render(Window1.boardCanvasStatic);

            replay_tick++;
        }
        //A shortcut function
        public static void RenderField(Canvas canvas, KeyValuePair<string, FieldType> field)
        {
            Board.RenderField(canvas, Board.STRtoXY(field.Key), field.Value);
        }
        //Renders a Field -> only thing to do is pick the color dependend on the field and render a circle
        //Plus adds events for Click on fields
        //Plus adds textboxes to show the players the String identifiers
        public static void RenderField(Canvas canvas, Point position, FieldType type)
        {
            Point hex_position = XYtoHex(position);
            Ellipse myEllipse = new Ellipse();

            SolidColorBrush mySolidColorBrush = new SolidColorBrush();
            mySolidColorBrush.Color = FieldFillColor(type);
            myEllipse.Fill = mySolidColorBrush;
            myEllipse.StrokeThickness = 2;
            myEllipse.Stroke = Brushes.Black;

            // Set the width and height of the Ellipse.
            myEllipse.Width = DIST / 1.2;
            myEllipse.Height = DIST / 1.2;

            //Adds the necessary data to identify the clicked position
            myEllipse.Name = Board.XYtoSTR(position);
            myEllipse.MouseDown += new MouseButtonEventHandler(Window1.myEllipse_MouseDown);


            // Add the Ellipse to the StackPanel.
            Canvas.SetLeft(myEllipse, hex_position.X);
            Canvas.SetBottom(myEllipse, hex_position.Y);
            canvas.Children.Add(myEllipse);

            TextBlock text_block = new TextBlock();
            text_block.Text = Board.XYtoSTR(position);
            text_block.Foreground = new SolidColorBrush(Color.FromRgb(0, 0, 0));

            //set the position of the text
            text_block.HorizontalAlignment = HorizontalAlignment.Center;
            text_block.VerticalAlignment = VerticalAlignment.Center;
            Canvas.SetLeft(text_block, hex_position.X + 10.0);
            Canvas.SetBottom(text_block, hex_position.Y + 10.0);

            //draw it to the canvas
            canvas.Children.Add(text_block);  
        }

        //For simplicity I redraw the whole screen. It not computionally complictated, 
        //because this happens no more then one time in a second.
        public void Render(Canvas canvas)
        {
            canvas.Children.Clear();
            foreach (KeyValuePair<string, FieldType> field in this.board)
            {
                Board.RenderField(canvas, field);
            }
        }

        //!TODO! -> disable moves backward
        //First I check all position where I can multijumb. Then I add the neighbour position when I can go
        public HashSet<Point> GetPossibleMoves(Point position)
        {
            #region VARIABLES
            //All positions where I can go
            HashSet<Point> result = new HashSet<Point>();
            //All directions in such Hex representation
          List<Point> go_thru = new List<Point>();
          go_thru.Add(new Point(0,1));  
          go_thru.Add(new Point(0,-1));
          go_thru.Add(new Point(1,0));
          go_thru.Add(new Point(-1,0));
          go_thru.Add(new Point(-1,1));
          go_thru.Add(new Point(1,-1));
            //buffer variables
          Point buf_pos1 = new Point();
          Point buf_pos2 = new Point();
            #endregion
          result.Add(new Point(position.X, position.Y));

          //move by two (multijump) - BFS
          while (true)
          {
              HashSet<Point> next_points = new HashSet<Point>();
              foreach (Point reached in result)
              {
                  foreach (Point move in go_thru)
                  {
           //Foreach position when I could jumped I checked if I can jump elsewhere
                      buf_pos1.X = reached.X + move.X;
                      buf_pos1.Y = reached.Y + move.Y;
                      buf_pos2.X = reached.X + 2*move.X;
                      buf_pos2.Y = reached.Y + 2*move.Y;
          //Buf1 -> the neighbour which I`m jumping over, Buf2 -> the field where I will land
                      //The next IF is quite self-descriptive
                      if (!result.Contains(buf_pos2) &&
                          this.board.ContainsKey(Board.XYtoSTR(buf_pos1)) &&
                          this.board.ContainsKey(Board.XYtoSTR(buf_pos2)) &&
                          this.board[Board.XYtoSTR(buf_pos1)] != FieldType.Empty &&
                          this.board[Board.XYtoSTR(buf_pos2)] == FieldType.Empty)
                      {
                          //If I can jump somewhere
                          next_points.Add(new Point(buf_pos2.X, buf_pos2.Y));
                      }
                  }
              }
              //Add the newly found points to all points
              result.UnionWith(next_points);

              //if no points found, break
              if (next_points.Count == 0) break;
              //break; //dev break //Development break
          }
          //remove the position where I`m standing (because I cannot go there)
          result.Remove(new Point(position.X, position.Y));

          //move by one
          foreach (Point move in go_thru)
          {
              buf_pos1.X = position.X + move.X;
              buf_pos1.Y = position.Y + move.Y;
              if (this.board.ContainsKey(Board.XYtoSTR(buf_pos1)) &&
                  this.board[Board.XYtoSTR(buf_pos1)] == FieldType.Empty)
              {
                  result.Add(new Point(buf_pos1.X, buf_pos1.Y));
              }
          }
          return result;
        }

        //Converters between representations of a position (the representation are 
        //descripted at the beginning of the Board class
        #region CONVERTERS
        static public Point XYtoHex(Point position)
        {
            return new Point(DIST * position.X + position.Y * DIST / 2.0, position.Y * DIST * Math.Sqrt(3.0) / 2.0);
        }
        static public Point STRtoXY(string s)
        {
            //return new Point(0, 0);
            return new Point((int)s[0] - (int)'A', Int32.Parse(s.Substring(1))); 
        }
        static public string XYtoSTR(Point position)
        {
            return ((char)(position.X + (int)'A')) + position.Y.ToString();  
        }
        static public Point STRtoHex(string s)
        {
            return XYtoHex(STRtoXY(s));
        }
        #endregion
        //Get the field color which depends only on the field type
        static public Color FieldFillColor(FieldType type)
        {
            switch (type)
            {
                case FieldType.Empty: return Color.FromRgb(255, 255, 255);
                case FieldType.Player1: return Color.FromRgb(200, 0, 0); 
                case FieldType.Player2: return Color.FromRgb(0, 200, 0); 
                case FieldType.Player3: return Color.FromRgb(0, 0, 200); 
                case FieldType.Player4: return Color.FromRgb(200, 200, 100); 
                case FieldType.Player5: return Color.FromRgb(200, 100, 200); 
                case FieldType.Player6: return Color.FromRgb(100, 200, 200); 
            }
            return Colors.White;
        }
    }
}
